import torch
from functools import wraps as fnwraps
from typing import Iterable, Mapping
from torch import nn
import torch.nn.functional as F
from .func_manager import FuncManager, FunctionModule
from .network_archs import NetNode, Drawer



def ver2num(version, vbits=2):
    version = version.split('.')
    number = 0
    for v in version:
        v = v.split('+')
        if len(v)>1:
            v = int(v[0], 16)
        else:
            v = int(v[0])
        number = number * 10 ** vbits + v
    return number

PT_VER = ver2num(torch.__version__)

class Inspector():
    def __init__(self, in_msg=(), out_msg=(), graph_path='./anetomy_inspector.png'):
        if graph_path.endswith('.png'):
            save_type = 'png'
        elif graph_path.endswith('.svg'):
            save_type = 'svg'
        else:
            raise NameError('NEBULAE ERROR ៙ graph path must be a png file or svg file.')
        if isinstance(in_msg, tuple):
            in_msg = in_msg + ('',)
        elif isinstance(in_msg, list):
            in_msg = in_msg + ['']
        if isinstance(out_msg, tuple):
            out_msg = out_msg + ('',)
        elif isinstance(out_msg, list):
            out_msg = out_msg + ['']
        assert len(in_msg) > 0 and len(out_msg) > 0
        self.tensors = []
        self.ops = []
        self.op_sn = {}
        self.fn_mng = FuncManager()
        self.drawer = Drawer(graph_path, save_type, in_msg, out_msg)

        def hook_func_old(func):
            @fnwraps(func)
            def wrapper(*args, **kwargs):
                fm = FunctionModule(func.__name__)
                if isinstance(args[0], (tuple, list)): # binary operators
                    flatten_input = tuple(args[0])+args[1:]+tuple(kwargs.values())
                else:
                    flatten_input = args+tuple(kwargs.values())
                hooked_input = self._before_forward(fm, flatten_input)
                ret = func(*args, **kwargs)
                self._after_forward(fm, hooked_input, ret)
                return ret
            return wrapper

        def hook_func_new(func):
            @fnwraps(func)
            def wrapper(*args, **kwargs):
                fm = FunctionModule(func.__name__)
                hooked_args, hooked_kwargs = self._before_forward(fm, args, kwargs)
                ret = func(*args, **kwargs)
                self._after_forward(fm, hooked_args, hooked_kwargs, ret)
                return ret
            return wrapper
        
        if PT_VER >= 2e4:
            self._hook_func = hook_func_new
            self._before_forward = self._before_forward_new
            self._after_forward = self._after_forward_new
        else:
            self._hook_func = hook_func_old
            self._before_forward = self._before_forward_old
            self._after_forward = self._after_forward_old
    
    def _hook(self, net):
        # avoid adding hooks one more time
        if not hasattr(net, '_record'):
            net._record = {'attr': None, 'times': 0}
            if PT_VER >= 2e4:
                net.register_forward_pre_hook(self._before_forward, with_kwargs=True)
                net.register_forward_hook(self._after_forward, with_kwargs=True)
            else:
                net.register_forward_pre_hook(self._before_forward)
                net.register_forward_hook(self._after_forward)
            
    def dissect(self, net, *dummy_args, **dummy_kwargs):
        self.fn_mng.decorate_funcs(self._hook_func)
        net.apply(self._hook)
        net(*dummy_args, **dummy_kwargs)
        for t in self.tensors:
            if t.last_op is not None:
                t.last_op.type = 'bim' if t.last_op.type=='bim' else 'leaf'
        self.fn_mng.undecorate_funcs()
        # clear invalid nodes
        for op in self.ops:
            if op.type in ('bim', 'leaf') and op.depth == 0:
                op.type = 'rot'
        self.entrance = [t for t in self.tensors if t.is_root() and (not t.is_invalid_pendant())]

    def draw(self, max_depth=-1, to_expand=()):
        self.drawer.draw(self.entrance, max_depth, to_expand)

    def _check_in_place(self, out_name, in_names):
        for name in in_names:
            op, stage = name.split('@')
            if stage.startswith('input') and op == out_name.split('@')[0]:
                return True
        return False
    
    def _arr2str(self, arr):
        s = []
        if isinstance(arr, Iterable):
            for a in arr:
                s.append('%d'%a)
            s = ', '.join(s)
        else:
            s = '%d'%arr
        return s
    
    def _record_builtin_attr(self, op):
        if isinstance(op, (nn.Conv1d, nn.Conv2d, nn.Conv3d)):
            self.curr_net.attrs = ['In Chs: %s'%self._arr2str(op.in_channels), 'Out Chs: %s'%self._arr2str(op.out_channels), \
                          'Kernel: %s'%self._arr2str(op.kernel_size), 'Stride: %s'%self._arr2str(op.stride), \
                          'Pad: %s'%self._arr2str(op.padding), 'Dilation: %s'%self._arr2str(op.dilation), \
                            'Group: %s'%self._arr2str(op.groups)]
            
        elif isinstance(op, (nn.ConvTranspose1d, nn.ConvTranspose2d, nn.ConvTranspose3d)):
            self.curr_net.attrs = ['In Chs: %s'%self._arr2str(op.in_channels), 'Out Chs: %s'%self._arr2str(op.out_channels), \
                          'Kernel: %s'%self._arr2str(op.kernel_size), 'Stride: %s'%self._arr2str(op.stride), \
                          'Pad: %s'%self._arr2str(op.padding), 'Out Pad: %s'%self._arr2str(op.output_padding), \
                            'Dilation: %s'%self._arr2str(op.dilation), 'Group: %s'%self._arr2str(op.groups)]
        elif isinstance(op, (nn.MaxPool1d, nn.MaxPool2d, nn.MaxPool3d, \
                             nn.AvgPool1d, nn.AvgPool2d, nn.AvgPool3d)):
            self.curr_net.attrs = ['Kernel: %s'%self._arr2str(op.kernel_size), \
                                   'Stride: %s'%self._arr2str(op.stride), \
                                    'Pad: %s'%self._arr2str(op.padding)]
        elif isinstance(op, nn.Linear):
            self.curr_net.attrs = ['In Chs: %s'%self._arr2str(op.in_features), \
                                   'Out Chs: %s'%self._arr2str(op.out_features)]
    
    def _return_or_create(self, tensor, name, label):
        # >| create a new node for tensors shown up the first time
        # >| duplicate those penetrating from root node
        # >| push them into an input or output list indexed by their depth
        is_in_place = False
        for t in self.tensors:
            for k, val in t.inputs_by_depth.items():
                for i, v in enumerate(val):
                    if tensor is v:
                        u = v.clone()
                        if self.curr_stage == 'in':
                            self.to_be_cloned.append((t, k+1, u, name))
                            self.to_be_linked.append((t, self.curr_net))
                            self.net_depth = max(self.net_depth, k+1)
                            return u
                        elif self.curr_stage == 'out': # only happens on in-place op
                            assert self._check_in_place(name, t.shared_names)
                            t.inputs_by_depth[k][i] = v.clone()
                            is_in_place = True
                            break
                        else:
                            raise KeyError('NEBULAE ERROR ៙ curr_stage must be either "in" or "out".')
                if is_in_place:
                    break
            for k, val in t.outputs_by_depth.items():
                for i, v in enumerate(val):
                    if tensor is v:
                        if self.curr_stage == 'in':
                            u = v.clone()
                            self.to_be_cloned.append((t, k, u, name))
                            self.to_be_linked.append((t, self.curr_net))
                            self.net_depth = k
                            return u
                        elif self.curr_stage == 'out':
                            if self._check_in_place(name, t.shared_names):
                                t.outputs_by_depth[k][i] = v.clone()
                                is_in_place = True
                                break
                            u = v.clone()
                            self.to_be_cloned.append((t, k-1, u, name))
                            self.to_be_linked.append((self.curr_net, t))
                            t.depth = min(t.depth, k-1)
                            t.name = name
                            t.label = label
                            return u
                        else:
                            raise KeyError('NEBULAE ERROR ៙ curr_stage must be either "in" or "out".')
                if is_in_place:
                    break
            if is_in_place:
                break
                        
        if self.curr_stage == 'in':
            self.to_be_cloned.append(NetNode(name, label, '#CC66FF', 0, tensor, self.curr_stage))
            self.to_be_linked.append(self.curr_net) # child
        elif self.curr_stage == 'out':
            self.to_be_cloned.append(NetNode(name, label, '#CC66FF', self.net_depth, tensor, self.curr_stage))
            self.to_be_linked.append(self.curr_net) # parent
        return tensor

    def _collect_tensors(self, tensors, name, label, idx=0):
        bucket = []
        nelem = 0
        if isinstance(tensors, torch.Tensor):
            struct = self._return_or_create(tensors, name + '@%sput_%d'%(self.curr_stage, idx), 
                                                    label + '@%sput_%d'%(self.curr_stage, idx))
            bucket.append(struct)
            nelem += 1
        elif isinstance(tensors, Mapping):
            struct = type(tensors)()
            for k, v in tensors.items():
                b, s, n = self._collect_tensors(v, name, label, idx)
                bucket.extend(b)
                struct[k] = s
                idx += n
                nelem += n
        elif isinstance(tensors, Iterable) and (not isinstance(tensors, (str, torch.Size))):
            struct = []
            for t in tensors:
                b, s, n = self._collect_tensors(t, name, label, idx)
                bucket.extend(b)
                struct.append(s)
                idx += n
                nelem += n
            struct = type(tensors)(struct)
        else:
            struct = tensors
        return bucket, struct, nelem

    def _before_forward_old(self, net, in_args):
        code = net.__name__ if hasattr(net, '__name__') else net.__class__.__name__
        # wrap input as tensors for arithmetic ops
        op_code = code.strip('_')
        if op_code.startswith('add') or op_code.startswith('iadd') or \
            op_code.startswith('sub') or op_code.startswith('isub') or \
            op_code.startswith('mul') or op_code.startswith('imul') or \
            op_code.startswith('div') or op_code.startswith('idiv'):
            list_in_args = []
            for ia in in_args:
                if not isinstance(ia, torch.Tensor):
                    list_in_args.append(torch.tensor(ia))
                else:
                    list_in_args.append(ia)
            in_args = tuple(list_in_args)
        # assign serial number to refrain from giving same names
        if code in self.op_sn.keys():
            self.op_sn[code] += 1
            name = code + '_%d' % self.op_sn[code]
            label = code + '_%d' % self.op_sn[code]
        else:
            self.op_sn[code] = 0
            name = code + '_0'
            label = code + '_0'
        self.net_depth = 0 # reset
        self.curr_stage = 'in'
        self.to_be_cloned = []
        self.to_be_linked = []
        self.curr_net = NetNode(name, label, '#FF3333', self.net_depth)
        # mark if this is a built-in module
        if self.fn_mng._belongs_to_builtin(net):
            self.curr_net.type = 'bim'
            if isinstance(net, FunctionModule):
                if net.__name__ in ('cat', 'stack'):
                    self.curr_net.attrs = ['Axis: %d'%(in_args[2])]
            else:
                self._record_builtin_attr(net)
            if not hasattr(self.curr_net, 'attrs'):
                self.curr_net.attrs = {'Attr: Null'}
        _, _in_args, m = self._collect_tensors(in_args, name, label)
        self.curr_net.depth = self.net_depth # input correct depth
        # hide it if this is a submodule of a built-in module
        is_visible = True
        tensor_buffer = []
        for i in range(len(self.to_be_cloned)):
            cloned = self.to_be_cloned[i]
            linked = self.to_be_linked[i]
            if isinstance(cloned, NetNode): # impossible to be hidden
                continue
            else:
                t, k, _, _ = cloned
                is_visible = t.inner_mark(self.curr_net, k)
                if not is_visible:
                    break
                tensor_buffer.append(t)
        if is_visible:
            for t in tensor_buffer:
                t.last_op = t.last_op_buffer
                t.nest_sep = t.nest_sep_buffer
            for i in range(len(self.to_be_cloned)):
                cloned = self.to_be_cloned[i]
                linked = self.to_be_linked[i]
                if isinstance(cloned, NetNode): # impossible to be hidden
                    self.tensors.append(cloned)
                    cloned.add_child(linked)
                    cloned.inner_mark(self.curr_net, 0)
                else:
                    t, k, u, n = cloned
                    if n not in t.shared_names:
                        t.shared_names.append(n)
                    if k not in t.inputs_by_depth.keys():
                        t.inputs_by_depth[k] = [u]
                    else:
                        t.inputs_by_depth[k].append(u)
                    linked[0].add_child(linked[1])
        else:
            del self.curr_net
            return in_args
        # if name.startswith('Conv2d') or name.startswith('conv2d'):
        #     import pdb;pdb.set_trace()
        if m == 0: # no tensors
            return in_args
        net.__dict__['_net_node'] = self.curr_net
        return _in_args

    def _after_forward_old(self, net, in_args, outs):
        self.curr_stage = 'out'
        self.to_be_cloned = []
        self.to_be_linked = []
        nnode = net.__dict__.get('_net_node', None)
        if nnode is None:
            return outs
        self.ops.append(nnode)
        self.curr_net = nnode
        self.net_depth = nnode.depth
        _, _outs, _ = self._collect_tensors(outs, nnode.name, nnode.label)
        for i in range(len(self.to_be_cloned)):
            cloned = self.to_be_cloned[i]
            linked = self.to_be_linked[i]
            if isinstance(cloned, NetNode):
                self.tensors.append(cloned)
                linked.add_child(cloned)
            else:
                t, k, u, n = cloned
                if n not in t.shared_names:
                    t.shared_names.append(n)
                if k not in t.outputs_by_depth.keys():
                    t.outputs_by_depth[k] = [u]
                else:
                    t.outputs_by_depth[k].append(u)
                linked[0].add_child(linked[1])
        net.__dict__.pop('_net_node')
        return _outs


    def _before_forward_new(self, net, in_args, in_kwargs):
        code = net.__name__ if hasattr(net, '__name__') else net.__class__.__name__
        # wrap input as tensors for arithmetic ops
        op_code = code.strip('_')
        if op_code.startswith('add') or op_code.startswith('iadd') or \
            op_code.startswith('sub') or op_code.startswith('isub') or \
            op_code.startswith('mul') or op_code.startswith('imul') or \
            op_code.startswith('div') or op_code.startswith('idiv'):
            list_in_args = []
            for ia in in_args:
                if not isinstance(ia, torch.Tensor):
                    list_in_args.append(torch.tensor(ia))
                else:
                    list_in_args.append(ia)
            in_args = tuple(list_in_args)
        # assign serial number to refrain from giving same names
        if code in self.op_sn.keys():
            self.op_sn[code] += 1
            name = code + '_%d' % self.op_sn[code]
            label = code + '_%d' % self.op_sn[code]
        else:
            self.op_sn[code] = 0
            name = code + '_0'
            label = code + '_0'
        self.net_depth = 0 # reset
        self.curr_stage = 'in'
        self.to_be_cloned = []
        self.to_be_linked = []
        self.curr_net = NetNode(name, label, '#FF3333', self.net_depth)
        # mark if this is a built-in module
        if self.fn_mng._belongs_to_builtin(net):
            self.curr_net.type = 'bim'
            if isinstance(net, FunctionModule):
                if net.__name__ in ('cat', 'stack'):
                    if len(in_args) > 1:
                        self.curr_net.attrs = ['Axis: %d'%(in_args[1])]
                    else:
                        self.curr_net.attrs = ['Axis: %d'%(in_kwargs['dim'])]
            else:
                self._record_builtin_attr(net)
            if not hasattr(self.curr_net, 'attrs'):
                self.curr_net.attrs = {'Attr: Null'}
        _, _in_args, m = self._collect_tensors(in_args, name, label)
        _, _in_kwargs, l = self._collect_tensors(in_kwargs, name, label, m)
        self.curr_net.depth = self.net_depth # input correct depth
        # hide it if this is a submodule of a built-in module
        is_visible = True
        tensor_buffer = []
        for i in range(len(self.to_be_cloned)):
            cloned = self.to_be_cloned[i]
            linked = self.to_be_linked[i]
            if isinstance(cloned, NetNode): # impossible to be hidden
                continue
            else:
                t, k, _, _ = cloned
                is_visible = t.inner_mark(self.curr_net, k)
                if not is_visible:
                    break
                tensor_buffer.append(t)
        if is_visible:
            for t in tensor_buffer:
                t.last_op = t.last_op_buffer
                t.nest_sep = t.nest_sep_buffer
            for i in range(len(self.to_be_cloned)):
                cloned = self.to_be_cloned[i]
                linked = self.to_be_linked[i]
                if isinstance(cloned, NetNode): # impossible to be hidden
                    self.tensors.append(cloned)
                    cloned.add_child(linked)
                    cloned.inner_mark(self.curr_net, 0)
                else:
                    t, k, u, n = cloned
                    if n not in t.shared_names:
                        t.shared_names.append(n)
                    if k not in t.inputs_by_depth.keys():
                        t.inputs_by_depth[k] = [u]
                    else:
                        t.inputs_by_depth[k].append(u)
                    linked[0].add_child(linked[1])
        else:
            del self.curr_net
            return in_args, in_kwargs
        if m+l == 0: # no tensors
            return in_args, in_kwargs
        net.__dict__['_net_node'] = self.curr_net
        return _in_args, _in_kwargs

    def _after_forward_new(self, net, in_args, in_kwargs, outs):
        self.curr_stage = 'out'
        self.to_be_cloned = []
        self.to_be_linked = []
        nnode = net.__dict__.get('_net_node', None)
        if nnode is None:
            return outs
        self.ops.append(nnode)
        self.curr_net = nnode
        self.net_depth = nnode.depth
        _, _outs, _ = self._collect_tensors(outs, nnode.name, nnode.label)
        for i in range(len(self.to_be_cloned)):
            cloned = self.to_be_cloned[i]
            linked = self.to_be_linked[i]
            if isinstance(cloned, NetNode):
                self.tensors.append(cloned)
                linked.add_child(cloned)
            else:
                t, k, u, n = cloned
                if n not in t.shared_names:
                    t.shared_names.append(n)
                if k not in t.outputs_by_depth.keys():
                    t.outputs_by_depth[k] = [u]
                else:
                    t.outputs_by_depth[k].append(u)
                linked[0].add_child(linked[1])
        net.__dict__.pop('_net_node')
        return _outs



if __name__ == '__main__':
    class BasicConv2d(nn.Module):
        def __init__(self, in_channels: int, out_channels: int, **kwargs) -> None:
            super().__init__()
            self.conv = nn.Conv2d(in_channels, out_channels, bias=False, **kwargs)
            self.bn = nn.BatchNorm2d(out_channels, eps=0.001, affine=False)
            self.relu = nn.ReLU()

        def forward(self, x):
            x = F.pad(x, [0,1,0,1])
            x = self.conv(x)
            y = self.bn(x)
            y = F.relu(y, inplace=True)
            return y
        
    class Dense(nn.Module):
        def __init__(self):
            super().__init__()
            self.lin = nn.Linear(28*28*2, 10)

        def forward(self, x):
            x = x.reshape(-1,)
            y = self.lin(x)
            return y

    class NeuralNetwork(nn.Module):
        def __init__(self):
            super().__init__()
            self.conv1 = nn.Conv2d(6, 8, 3, padding=1)
            self.mpool1 = nn.MaxPool2d(3, stride=2, ceil_mode=True)
            self.conv2 = BasicConv2d(8, 8, kernel_size=3, stride=2, padding=0)
            self.dense = Dense()

        def forward(self, x, scale=2):
            x -= scale
            z = x*scale
            y = torch.cat([x, z], dim=1)
            y = self.conv1(y)
            y = self.mpool1(y)
            y = self.conv2(y)
            y = self.dense(y)
            return y


    # ------------------ main ------------------- #
    net = NeuralNetwork()
    dummy_x = torch.randn(1, 3, 56, 56)
    dummy_s = 2
    ant = Inspector(('Image',), ('Helmet: 82%',))
    ant.dissect(net, dummy_x, scale=dummy_s)
    ant.draw(3)


import os
from graphviz import Digraph


class NetNode():
    def __init__(self, name, label, color, depth, tensor=None, type='op'):
        self.name = name
        self.label = label
        self.color = color
        self.depth = depth
        self.type = type
        if type == 'in':
            self.inputs_by_depth = {depth: [tensor]}
            self.outputs_by_depth = {}
            self.shape = tensor.shape
        elif type == 'out':
            self.inputs_by_depth = {}
            self.outputs_by_depth = {depth: [tensor]}
            self.shape = tensor.shape
        self.shared_names = [name]
        self.last_op = None
        self.nest_sep = []
        # save changes in buffer until inner_mark is done in case of multiple input
        self.last_op_buffer = None
        self.nest_sep_buffer = []
        self.prev = []
        self.next = []

    def add_child(self, child, log=False):
        if log:
            print(self.name, '#%d'%(self.depth), '->', child.name, '#%d'%(child.depth))
        if child not in self.next:
            self.next.append(child)
        # complete upstream info
        if self not in child.prev:
            child.prev.append(self)

    def inner_mark(self, curr_mod, in_depth):
        # |> mark it if the last module is a leaf module
        # |> note that it doesn't mean this is a leaf node
        if self.last_op is None:
            self.last_op_buffer = curr_mod
        else:
            if in_depth <= self.nest_sep[-1]:
                self.last_op.type = 'bim' if self.last_op.type=='bim' else 'leaf'
            elif self.last_op.type == 'bim':
                return False
            self.last_op_buffer = curr_mod
        self.nest_sep_buffer.append(in_depth)
        return True

    def is_root(self):
        return len(self.prev) == 0
    
    def is_leaf(self):
        return len(self.next) == 0
    
    def is_invalid_pendant(self):
        invalid = True
        if self.is_root():
            for n in self.next:
                if n.type != 'rot':
                    invalid = False
                    break
        if self.is_leaf():
            for p in self.prev:
                if p.type != 'rot':
                    invalid = False
                    break
        return invalid



class Drawer():
    def __init__(self, graph_path, save_type, in_msg, out_msg):
        self.graph_path = graph_path[:-4]
        self.save_type = save_type
        self.in_msg = in_msg
        self.out_msg = out_msg
        self.graph = Digraph(name='AINN', graph_attr={'fontname': "Helvetica,Arial,sans-serif"})

    def draw(self, entrance, max_depth, to_expand, in_data=None):
        # >| -1:  expand all layers
        # >| N:  collapse layers that are deeper than N or in the input list
        assert isinstance(to_expand, (list, tuple, str)), \
            'NEBULAE ERROR ៙ the modules to be expanded must be passed in as an array or string.'
        self.max_depth = max_depth if max_depth >= 0 else float('inf')
        self.seen = [] # whether has been created
        self.expanded = [] # whether has been expanded
        self.to_expand = list(to_expand) if isinstance(to_expand, (list, tuple)) else [to_expand]
        self.graph_stack = [self.graph]
        self.node_stack = []
        if in_data is None:
            in_data = len(entrance) * [None]
        assert len(in_data) == len(entrance)
        
        # >| starts from entrance nodes
        self.out_idx = 0
        for i, ent in enumerate(entrance):
            # if (os.path.dirname(ent.name) not in self.to_expand) or (ent.name in self.to_expand):
            #     continue
            message = self.in_msg[min(i, len(self.in_msg)-1)]
            self.graph.node(ent.name, f"""<
                            <table border="0" cellborder="0" cellpadding="3" bgcolor="white">
                                <tr>
                                    <td bgcolor="black" align="center" colspan="2"><font color="white">{ent.label} D:0</font></td>
                                </tr>
                                <tr>
                                    <td align="left" port="r5">{message}</td>
                                </tr>
                            </table>
                            >""",
                            color=ent.color, penwidth="2", 
                            style="filled", fillcolor="white", 
                            fontname="Courier New", shape="box")
            if in_data[i] is not None:
                self.graph.node(ent.name + '_viz', ' ',
                                color=ent.color, penwidth="2", 
                                style="filled", fillcolor="white", 
                                fontname="Courier New", shape="box",
                                image=in_data[i], imagepos='mc', 
                                imagescale='true', width='1.1', height='1.1', fixedsize='true')
                self.graph.edge(ent.name, ent.name + '_viz', constraint='false')
            self.seen.append(ent.name)
            self.node_stack.append([ent.name])
            self.within_subg = 1
            self._expand(ent)
        self.graph.render(self.graph_path, view=False, format=self.save_type)
        os.remove(self.graph_path)

    def _expand(self, root, stack_in=True):
        if root.name in self.expanded:
            return
        if stack_in:
            self.node_stack.append([])
        for c in root.next:
            # print(root.shared_names, '==>', c.shared_names)
            # import pdb;pdb.set_trace()
            if c.type == 'out':
                assert root.type in ('op', 'bim', 'leaf')
                for name in c.shared_names:
                    subg_name = self.graph_stack[-1].name[8:]
                    if name.startswith(subg_name):
                        if len(self.graph_stack) > 1:
                            self.graph_stack[-2].subgraph(self.graph_stack[-1])
                            del self.graph_stack[-1]
                if c.is_leaf():
                    if c.name not in self.seen:
                        if c.depth == 0:
                            message = self.out_msg[min(self.out_idx, len(self.out_msg)-1)]
                            self.out_idx += 1
                        else:
                            message = ''
                        self.graph_stack[-1].node(c.name, f"""<
                                <table border="0" cellborder="0" cellpadding="3" bgcolor="white">
                                    <tr>
                                        <td bgcolor="black" align="center" colspan="2"><font color="white">{c.label} D:{c.depth}</font></td>
                                    </tr>
                                    <tr>
                                        <td align="left" port="r5">{message}</td>
                                    </tr>
                                </table>
                                >""",
                                color=c.color, penwidth="2", 
                                style="filled", fillcolor="white", 
                                fontname="Courier New", shape="Mrecord")
                        self.seen.append(c.name)
                    node_name = self.node_stack[-2] # get last layer
                    while not isinstance(node_name, str):
                        node_name = node_name[-1]
                    subg = self.graph_stack[-self.within_subg]
                    self.within_subg = 1
                    subg.edge(node_name, c.name, ' x '.join([str(s) for s in c.shape]))
                else:
                    self._expand(c, False)
            elif c.type == 'in':
                raise KeyError
            elif (c.type in ('bim', 'leaf') and c.depth <= self.max_depth) or (c.type == 'op' and c.depth == self.max_depth):
                assert root.type in ('in', 'out')
                if c.name not in self.seen:
                    # import pdb; pdb.set_trace()
                    if c.type == 'bim':
                        message = '\n'.join(['<tr><td align="left" port="r5">%s</td></tr>'%a for a in c.attrs])
                    else:
                        message = '<tr><td align="left" port="r5">Attr: Null</td></tr>'
                    self.graph_stack[-1].node(c.name, f"""<
                                <table border="0" cellborder="0" cellpadding="3" bgcolor="white">
                                    <tr>
                                        <td bgcolor="black" align="center" colspan="2"><font color="white">{c.label} D:{c.depth}</font></td>
                                    </tr>
                                    {message}
                                </table>
                                >""",
                                color=c.color, penwidth="2", 
                                style="filled", fillcolor="white", 
                                fontname="Courier New", shape="Mrecord")
                    self.seen.append(c.name)
                node_name = self.node_stack[-2] # get last layer
                while not isinstance(node_name, str):
                    node_name = node_name[-1]
                subg = self.graph_stack[-self.within_subg]
                self.within_subg = 1
                subg.edge(node_name, c.name, ' x '.join([str(s) for s in root.shape]))
                self.node_stack[-1].append(c.name)
                self._expand(c)
            elif (c.type == 'op' and c.depth < self.max_depth) or c.name in self.to_expand:
                assert root.type in ('in', 'out')
                if c.name not in self.seen:
                    subg = Digraph(name='cluster_%s'%c.name, graph_attr={'fontname': "Helvetica,Arial,sans-serif"})
                    subg.attr(style='dashed', color='teal', label=c.label, margin='20 30 20 30') # Top, Right, Bottom, Left
                    self.graph_stack.append(subg)
                    self.within_subg += 1
                self.seen.append(c.name) # only record but not to expand cuz the input will penetrate in
            elif c.type in ('op', 'bim', 'leaf') and c.depth > self.max_depth:
                continue
            else:
                raise TypeError('NEBULAE ERROR ៙ current node %s is rotten or recorded inproperly.'%c.name)
        self.expanded.append(root.name)
        if stack_in:
            del self.node_stack[-1]